<?php
// Own namespace
namespace CoreFramework\Filesytem\Pointer;

// Import framework stuff
use CoreFramework\Object\BaseFrameworkSystem;

/**
 * A class for directory reading and getting its contents, no recursion!
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2017 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class FrameworkDirectoryPointer extends BaseFrameworkSystem implements FrameworkDirectory {
	/**
	 * The current path we are working in
	 */
	private $pathName = '';

	/**
	 * The directory iterator instance
	 */
	private $iteratorInstance = NULL;

	/**
	 * Protected constructor
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Destructor for cleaning purposes, etc
	 */
	public function __destruct() {
		// Is there a resource pointer? Then we have to close the directory here!
		if ($this->getDirectoryIteratorInstance() instanceof DirectoryIterator) {
			// Try to close a directory
			$this->closeDirectory();
		} // END - if

		// Call the parent destructor
		parent::__destruct();
	}

	/**
	 * Create a directory pointer based on the given path. The path will also
	 * be verified here.
	 *
	 * @param	$pathName		The path name we shall pass to opendir()
	 * @param	$inConstructor	If we are in de/con-structor or from somewhere else
	 * @return	$pointerInstance	A prepared instance of FrameworkDirectoryPointer
	 * @throws	PathIsEmptyException	If the provided path name is empty
	 * @throws	InvalidPathStringException	If the provided path name is not a string
	 * @throws	PathIsNoDirectoryException	If the provided path name is not valid
	 * @throws	PathReadProtectedException	If the provided path name is read-protected
	 * @todo	Get rid of inConstructor, could be old-lost code.
	 */
	public static final function createFrameworkDirectoryPointer ($pathName, $inConstructor = FALSE) {
		// Some pre-sanity checks...
		if (is_null($pathName)) {
			// No pathname given
			if ($inConstructor) {
				return NULL;
			} else {
				throw new PathIsEmptyException(NULL, self::EXCEPTION_UNEXPECTED_EMPTY_STRING);
			}
		} elseif (!is_string($pathName)) {
			// Is not a string
			if ($inConstructor) {
				return NULL;
			} else {
				throw new InvalidPathStringException(NULL, self::EXCEPTION_INVALID_STRING);
			}
		} elseif (!is_dir($pathName)) {
			// Not a directory
			if ($inConstructor) {
				return NULL;
			} else {
				throw new PathIsNoDirectoryException($pathName, self::EXCEPTION_INVALID_PATH_NAME);
			}
		} elseif (!is_readable($pathName)) {
			// Not readable
			if ($inConstructor) {
				return NULL;
			} else {
				throw new PathReadProtectedException($pathName, self::EXCEPTION_READ_PROTECED_PATH);
			}
		}

		// Create new instance
		$pointerInstance = new FrameworkDirectoryPointer();

		// Get an iterator for the directory
		$iteratorInstance = new DirectoryIterator($pathName);

		// ... and rewind back
		$iteratorInstance->rewind();

		// Set directory pointer and path name
		$pointerInstance->setDirectoryIteratorInstance($iteratorInstance);
		$pointerInstance->setPathName($pathName);

		// Return the instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: Opened pathName=' . $pathName . ' - EXIT!');
		return $pointerInstance;
	}

	/**
	 * Read raw lines of data from a directory pointer and return the data
	 *
	 * @return	$currentEntry	Current entry from encapsulated iterator
	 */
	public function readRawDirectory () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . '] - CALLED!');

		// Can the next entry be read?
		assert($this->getDirectoryIteratorInstance()->valid());

		// Read data from the directory pointer and return it
		$currentEntry = $this->getDirectoryIteratorInstance()->current();

		// Return found entry
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: currentEntry[]=' . gettype($currentEntry) . ' - EXIT!');
		return $currentEntry;
	}

	/**
	 * Read lines from the current directory pointer except some parts
	 *
	 * @param	$except		Some parts of a directory we want to ignore. Valid: directory and file names, other values will be silently ignored
	 * @return	string		Directory and/or file names read from the current directory pointer
	 */
	public function readDirectoryExcept (array $except = array()) {
		// No exceptions given?
		if (count($except) == 0) {
			// No exception given, so read all files and directories, but not recursive
			self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: No exceptions given, please use readRawDirectory() instead!');
			return $this->readRawDirectory();
		} elseif (!$this->getDirectoryIteratorInstance()->valid()) {
			// No more left to read
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: EOD reached.');
			return NULL;
		}

		// Init raw line
		$rawLine = NULL;

		// Read a raw line...
		$currentEntry = $this->readRawDirectory();
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: currentEntry[]=' . gettype($currentEntry));

		// Shall we exclude directories?
		if (is_object($currentEntry)) {
			// Get file name
			$rawLine = $currentEntry->getFilename();
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: rawLine[' . gettype($rawLine) . ']=' . $rawLine . ',isDot=' . intval($this->getDirectoryIteratorInstance()->isDot()));

			// Is it a dot-directory or excluded?
			if (($this->getDirectoryIteratorInstance()->isDot()) || (in_array($rawLine, $except))) {
				// To next entry
				$this->getDirectoryIteratorInstance()->next();

				// Exclude this part
				$rawLine = $this->readDirectoryExcept($except);
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: rawline[' . gettype($rawLine) . ']=' . $rawLine . ' - Recursive call!');
			} // END - if
		} // END - if

		// To next entry
		$this->getDirectoryIteratorInstance()->next();

		// Return read line
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('DIRECTORY[' . __METHOD__ . ':' . __LINE__ . ']: rawLine[' . gettype($rawLine) . ']=' . $rawLine);
		return $rawLine;
	}

	/**
	 * Close a directory source and set it's instance to null and the path name
	 * to empty
	 *
	 * @return	void
	 */
	public function closeDirectory () {
		// Close the directory by unsetting it
		$this->unsetDirectoryIteratorInstance();
		$this->setPathName('');
	}

	/**
	 * Setter for the directory pointer
	 *
	 * @param	$iteratorInstance	An instanceof a DirectoryIterator class
	 * @return	void
	 */
	protected final function setDirectoryIteratorInstance (DirectoryIterator $iteratorInstance) {
		// Set instance
		$this->iteratorInstance = $iteratorInstance;
	}

	/**
	 * Getter for the directory pointer
	 *
	 * @return	$iteratorInstance	The directory pointer which shall be a valid directory resource
	 */
	public final function getDirectoryIteratorInstance () {
		return $this->iteratorInstance;
	}

	/**
	 * Remove directory iterator instance (effectively closing it) by setting
	 * it to NULL. This will trigger a call on the destructor which will then
	 * "close" the iterator.
	 *
	 * @param	$iteratorInstance	An instanceof a DirectoryIterator class
	 * @return	void
	 */
	protected final function unsetDirectoryIteratorInstance () {
		// "Unset" the instance
		$this->iteratorInstance = NULL;
	}

	/**
	 * Setter for path name
	 *
	 * @param	$pathName	The new path name
	 * @return	void
	 */
	protected final function setPathName ($pathName) {
		$pathName = (string) $pathName;
		$this->pathName = $pathName;
	}

	/**
	 * Getter for path name
	 *
	 * @return	$pathName	The current path name
	 */
	public final function getPathName () {
		return $this->pathName;
	}

}
