<?php
// Own namespace
namespace CoreFramework\Output;

// Import framework stuff
use CoreFramework\Manager\ManageableApplication;
use CoreFramework\Registry\Registerable;

/**
 * This class simply puts HTML code / JavaScript code or CSS code out to the
 * browser
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2017 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class WebOutput extends BaseOutput implements OutputStreamer, Registerable {
	/**
	 * The instance for the singleton design pattern
	 */
	private static $webInstance = NULL;

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Create a new web output system and set the content type
	 *
	 * @param	$applicationInstance	An instance of a ManageableApplication class
	 * @return	$debugInstance			An instance of this middleware class
	 */
	public static final function createWebOutput (ManageableApplication $applicationInstance) {
		// Is the self-instance already set?
		if (is_null(self::$webInstance)) {
			// Get a new instance and set it
			self::$webInstance = new WebOutput();

			// Get the content type
			$contentType = self::$webInstance->getConfigInstance()->getConfigEntry('web_content_type');

			// Set the content type
			if (!empty($contentType)) {
				// Set the header
				$applicationInstance->getResponseInstance()->addHeader('Content-type', $contentType);
			} // END - if
		} // END - if

		// Return instance
		return self::$webInstance;
	}

	/**
	 * Output the code
	 *
	 * @param	$outStream	Stream to output
	 * @param	$stripTags	Whether HTML tags shall be stripped out
	 * @return	void
	 */
	public final function output ($outStream = FALSE, $stripTags = FALSE) {
		print(stripslashes($outStream));
	}

	/**
	 * Determines seek position
	 *
	 * @return	$seekPosition	Current seek position
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function determineSeekPosition () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Seek to given offset (default) or other possibilities as fseek() gives.
	 *
	 * @param	$offset		Offset to seek to (or used as "base" for other seeks)
	 * @param	$whence		Added to offset (default: only use offset to seek to)
	 * @return	$status		Status of file seek: 0 = success, -1 = failed
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function seek ($offset, $whence = SEEK_SET) {
		self::createDebugInstance(__CLASS__)->debugOutput('[' . __METHOD__ . ':' . __LINE__ . '] offset=' . $offset . ',whence=' . $whence);
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Size of file stack
	 *
	 * @return	$size	Size (in bytes) of file
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function size () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}
}

// [EOF]
?>
