<?php
/**
 * A general file-based stack class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2017 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
class BaseFileStack extends BaseStacker {
	/**
	 * Magic for this stack
	 */
	const STACK_MAGIC = 'STACKv0.1';

	/**
	 * Name of array index for gap position
	 */
	const ARRAY_INDEX_GAP_POSITION = 'gap';

	/**
	 * Name of array index for hash
	 */
	const ARRAY_INDEX_HASH = 'hash';

	/**
	 * Name of array index for length of raw data
	 */
	const ARRAY_INDEX_DATA_LENGTH = 'length';

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct ($className) {
		// Call parent constructor
		parent::__construct($className);
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 * @todo	To hard assertions here, better rewrite them to exceptions
	 */
	public function readFileHeader () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// First rewind to beginning as the header sits at the beginning ...
		$this->getIteratorInstance()->rewind();

		// Then read it (see constructor for calculation)
		$data = $this->getIteratorInstance()->read($this->getIteratorInstance()->getHeaderSize());
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Read %d bytes (%d wanted).', __METHOD__, __LINE__, strlen($data), $this->getIteratorInstance()->getHeaderSize()));

		// Have all requested bytes been read?
		assert(strlen($data) == $this->getIteratorInstance()->getHeaderSize());
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Last character must be the separator
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] data(-1)=%s', __METHOD__, __LINE__, dechex(ord(substr($data, -1, 1)))));
		assert(substr($data, -1, 1) == chr(BaseBinaryFile::SEPARATOR_HEADER_ENTRIES));
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Okay, then remove it
		$data = substr($data, 0, -1);

		// And update seek position
		$this->getIteratorInstance()->updateSeekPosition();

		/*
		 * Now split it:
		 *
		 * 0 => magic
		 * 1 => total entries
		 * 2 => current seek position
		 */
		$header = explode(chr(BaseBinaryFile::SEPARATOR_HEADER_DATA), $data);

		// Set header here
		$this->getIteratorInstance()->setHeader($header);

		// Check if the array has only 3 elements
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] header(%d)=%s', __METHOD__, __LINE__, count($header), print_r($header, TRUE)));
		assert(count($header) == 3);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Check magic
		assert($header[0] == self::STACK_MAGIC);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Check length of count and seek position
		assert(strlen($header[1]) == BaseBinaryFile::LENGTH_COUNT);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));
		assert(strlen($header[2]) == BaseBinaryFile::LENGTH_POSITION);
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Passed assert().', __METHOD__, __LINE__));

		// Decode count and seek position
		$header[1] = hex2bin($header[1]);
		$header[2] = hex2bin($header[2]);

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 */
	public function flushFileHeader () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// Put all informations together
		$header = sprintf('%s%s%s%s%s%s',
			// Magic
			self::STACK_MAGIC,

			// Separator magic<->count
			chr(BaseBinaryFile::SEPARATOR_HEADER_DATA),

			// Total entries (will be zero) and pad it to 20 chars
			str_pad($this->dec2hex($this->getIteratorInstance()->getCounter()), BaseBinaryFile::LENGTH_COUNT, '0', STR_PAD_LEFT),

			// Separator count<->seek position
			chr(BaseBinaryFile::SEPARATOR_HEADER_DATA),

			// Position (will be zero)
			str_pad($this->dec2hex($this->getIteratorInstance()->getSeekPosition(), 2), BaseBinaryFile::LENGTH_POSITION, '0', STR_PAD_LEFT),

			// Separator position<->entries
			chr(BaseBinaryFile::SEPARATOR_HEADER_ENTRIES)
		);

		// Write it to disk (header is always at seek position 0)
		$this->getIteratorInstance()->writeData(0, $header, FALSE);

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Initializes this file-based stack.
	 *
	 * @param	$fileName	File name of this stack
	 * @param	$type		Type of this stack (e.g. url_source for URL sources)
	 * @return	void
	 * @todo	Currently the stack file is not cached, please implement a memory-handling class and if enough RAM is found, cache the whole stack file.
	 */
	protected function initFileStack ($fileName, $type) {
		// Get a stack file instance
		$fileInstance = ObjectFactory::createObjectByConfiguredName('stack_file_class', array($fileName, $this));

		// Get iterator instance
		$iteratorInstance = ObjectFactory::createObjectByConfiguredName('file_iterator_class', array($fileInstance));

		// Is the instance implementing the right interface?
		assert($iteratorInstance instanceof SeekableWritableFileIterator);

		// Set iterator here
		$this->setIteratorInstance($iteratorInstance);

		// Calculate header size
		$this->getIteratorInstance()->setHeaderSize(
			strlen(self::STACK_MAGIC) +
			strlen(chr(BaseBinaryFile::SEPARATOR_HEADER_DATA)) +
			BaseBinaryFile::LENGTH_COUNT +
			strlen(chr(BaseBinaryFile::SEPARATOR_HEADER_DATA)) +
			BaseBinaryFile::LENGTH_POSITION +
			strlen(chr(BaseBinaryFile::SEPARATOR_HEADER_ENTRIES))
		);

		// Init counters and gaps array
		$this->getIteratorInstance()->initCountersGapsArray();

		// Is the file's header initialized?
		if (!$this->getIteratorInstance()->isFileHeaderInitialized()) {
			// No, then create it (which may pre-allocate the stack)
			$this->getIteratorInstance()->createFileHeader();

			// And pre-allocate a bit
			$this->getIteratorInstance()->preAllocateFile('file_stack');
		} // END - if

		// Load the file header
		$this->readFileHeader();

		// Count all entries in file
		$this->getIteratorInstance()->analyzeFile();

		/*
		 * Get stack index instance. This can be used for faster
		 * "defragmentation" and startup.
		 */
		$indexInstance = FileStackIndexFactory::createFileStackIndexInstance($fileName, $type);

		// And set it here
		$this->setIndexInstance($indexInstance);
	}

	/**
	 * Adds a value to given stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @param	$value			Value to add to this stacker
	 * @return	void
	 * @throws	FullStackerException	If the stack is full
	 */
	protected function addValue ($stackerName, $value) {
		// Do some tests
		if ($this->isStackFull($stackerName)) {
			// Stacker is full
			throw new FullStackerException(array($this, $stackerName, $value), self::EXCEPTION_STACKER_IS_FULL);
		} // END - if

		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName . ',value[' . gettype($value) . ']=' . print_r($value, TRUE));

		// No objects/resources are allowed as their serialization takes to long
		assert(!is_object($value));
		assert(!is_resource($value));

		/*
		 * Now add the value to the file stack which returns gap position, a
		 * hash and length of the raw data.
		 */
		$data = $this->getIteratorInstance()->writeValueToFile($stackerName, $value);

		// Add the hash and gap position to the index
		$this->getIndexInstance()->addHashToIndex($stackerName, $data);
	}

	/**
	 * Get last value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	EmptyStackerException	If the stack is empty
	 */
	protected function getLastValue ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now get the last value
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		return $value;
	}

	/**
	 * Get first value from named stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value of last added value
	 * @throws	EmptyStackerException	If the stack is empty
	 */
	protected function getFirstValue ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now get the first value
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		$value = NULL;

		// Return it
		return $value;
	}

	/**
	 * "Pops" last entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "poped" from array
	 * @throws	EmptyStackerException	If the stack is empty
	 */
	protected function popLast ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		return NULL;
	}

	/**
	 * "Pops" first entry from stack
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$value			Value "shifted" from array
	 * @throws	EmptyStackerException	If the named stacker is empty
	 */
	protected function popFirst ($stackerName) {
		// Is the stack not yet initialized or full?
		if ($this->isStackEmpty($stackerName)) {
			// Throw an exception
			throw new EmptyStackerException(array($this, $stackerName), self::EXCEPTION_STACKER_IS_EMPTY);
		} // END - if

		// Now, remove the last entry, we don't care about the return value here, see elseif() block above
		/* NOISY-DEBUG: */ $this->partialStub('[' . __METHOD__ . ':' . __LINE__ . '] stackerName=' . $stackerName);
		return NULL;
	}

	/**
	 * Checks whether the given stack is full
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$isFull			Whether the stack is full
	 */
	protected function isStackFull ($stackerName) {
		// File-based stacks will only run full if the disk space is low.
		// @TODO Please implement this, returning FALSE
		$isFull = FALSE;

		// Return result
		return $isFull;
	}

	/**
	 * Checks whether the given stack is empty
	 *
	 * @param	$stackerName		Name of the stack
	 * @return	$isEmpty			Whether the stack is empty
	 * @throws	NoStackerException	If given stack is missing
	 */
	public function isStackEmpty ($stackerName) {
		// So, is the stack empty?
		$isEmpty = (($this->getStackCount($stackerName)) == 0);

		// Return result
		return $isEmpty;
	}

	/**
	 * Initializes given stacker
	 *
	 * @param	$stackerName	Name of the stack
	 * @param	$forceReInit	Force re-initialization
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function initStack ($stackerName, $forceReInit = FALSE) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Initializes all stacks
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function initStacks (array $stacks, $forceReInit = FALSE) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Checks whether the given stack is initialized (set in array $stackers)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$isInitialized	Whether the stack is initialized
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function isStackInitialized ($stackerName) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Determines whether the EOF has been reached
	 *
	 * @return	$isEndOfFileReached		Whether the EOF has been reached
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function isEndOfFileReached () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for file name
	 *
	 * @return	$fileName	The current file name
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function getFileName () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for size of given stack (array count)
	 *
	 * @param	$stackerName	Name of the stack
	 * @return	$count			Size of stack (array count)
	 */
	public function getStackCount ($stackerName) {
		// Now, simply return the found count value, this must be up-to-date then!
		return $this->getIteratorInstance()->getCounter();
	}

	/**
	 * Calculates minimum length for one entry/block
	 *
	 * @return	$length		Minimum length for one entry/block
	 */
	public function calculateMinimumBlockLength () {
		// Calulcate it
		$length =
			// Length of entry group
			BaseBinaryFile::LENGTH_GROUP + strlen(chr(BaseBinaryFile::SEPARATOR_GROUP_HASH)) +
			// Hash + value
			self::getHashLength() + strlen(chr(BaseBinaryFile::SEPARATOR_HASH_VALUE)) + 1 +
			// Final separator
			strlen(chr(BaseBinaryFile::SEPARATOR_ENTRIES));

		// Return it
		return $length;
	}

	/**
	 * Initializes counter for valid entries, arrays for damaged entries and
	 * an array for gap seek positions. If you call this method on your own,
	 * please re-analyze the file structure. So you are better to call
	 * analyzeFile() instead of this method.
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function initCountersGapsArray () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for header size
	 *
	 * @return	$totalEntries	Size of file header
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function getHeaderSize () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Setter for header size
	 *
	 * @param	$headerSize		Size of file header
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function setHeaderSize ($headerSize) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for header array
	 *
	 * @return	$totalEntries	Size of file header
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function getHeader () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Setter for header
	 *
	 * @param	$header		Array for a file header
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function setHeader (array $header) {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Updates seekPosition attribute from file to avoid to much access on file.
	 *
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function updateSeekPosition () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Getter for total entries
	 *
	 * @return	$totalEntries	Total entries in this file
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public final function getCounter () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Writes data at given position
	 *
	 * @param	$seekPosition	Seek position
	 * @param	$data			Data to be written
	 * @param	$flushHeader	Whether to flush the header (default: flush)
	 * @return	void
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function writeData ($seekPosition, $data, $flushHeader = TRUE) {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] seekPosition=%s,data[]=%s,flushHeader=%d', __METHOD__, __LINE__, $seekPosition, gettype($data), intval($flushHeader)));
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Writes given value to the file and returns a hash and gap position for it
	 *
	 * @param	$groupId	Group identifier
	 * @param	$value		Value to be added to the stack
	 * @return	$data		Hash and gap position
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function writeValueToFile ($groupId, $value) {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] groupId=%s,value[%s]=%s', __METHOD__, __LINE__, $groupId, gettype($value), print_r($value, TRUE)));
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Searches for next suitable gap the given length of data can fit in
	 * including padding bytes.
	 *
	 * @param	$length			Length of raw data
	 * @return	$seekPosition	Found next gap's seek position
	 * @throws	UnsupportedOperationException	This method is not (and maybe never will be) supported
	 */
	public function searchNextGap ($length) {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] length=%s', __METHOD__, __LINE__, $length));
		throw new UnsupportedOperationException(array($this, __FUNCTION__, $this->getIteratorInstance()->getPointerInstance()), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * "Getter" for file size
	 *
	 * @return	$fileSize	Size of currently loaded file
	 */
	public function getFileSize () {
		// Call iterator's method
		return $this->getIteratorInstance()->getFileSize();
	}

	/**
	 * Writes given raw data to the file and returns a gap position and length
	 *
	 * @param	$groupId	Group identifier
	 * @param	$hash		Hash from encoded value
	 * @param	$encoded	Encoded value to be written to the file
	 * @return	$data		Gap position and length of the raw data
	 */
	public function writeDataToFreeGap ($groupId, $hash, $encoded) {
		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,encoded()=%d - CALLED!', __METHOD__, __LINE__, $groupId, $hash, strlen($encoded)));

		// Raw data been written to the file
		$rawData = sprintf('%s%s%s%s%s',
			$groupId,
			BaseBinaryFile::SEPARATOR_GROUP_HASH,
			hex2bin($hash),
			BaseBinaryFile::SEPARATOR_HASH_VALUE,
			$encoded
		);

		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,rawData()=%d', __METHOD__, __LINE__, $groupId, $hash, strlen($rawData)));

		// Search for next free gap
		$gapPosition = $this->getIteratorInstance()->searchNextGap(strlen($rawData));

		// Gap position cannot be smaller than header length + 1
		assert($gapPosition > $this->getIteratorInstance()->getHeaderSize());

		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,gapPosition=%s', __METHOD__, __LINE__, $groupId, $hash, $gapPosition));

		// Then write the data at that gap
		$this->getIteratorInstance()->writeData($gapPosition, $rawData);

		// Debug message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] groupId=%s,hash=%s,rawData()=%d - EXIT!', __METHOD__, __LINE__, $groupId, $hash, strlen($rawData)));

		// Return gap position, hash and length of raw data
		return array(
			self::ARRAY_INDEX_GAP_POSITION => $gapPosition,
			self::ARRAY_INDEX_HASH         => $hash,
			self::ARRAY_INDEX_DATA_LENGTH  => strlen($rawData)
		);
	}
}

// [EOF]
?>
